import { __awaiter } from 'tslib';
import { FormArray, FormBuilder, FormControl, FormGroup } from '@angular/forms';
import { classToClass, plainToClass } from 'class-transformer';
import { getFromContainer, MetadataStorage, validate, validateSync, ValidationTypes, Validator } from 'class-validator';
import { cloneDeep, mergeWith } from 'lodash-es';
import 'reflect-metadata';
import { BehaviorSubject } from 'rxjs';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template TModel
 */
class DynamicFormGroup extends FormGroup {
    /**
     * @param {?} factoryModel
     * @param {?} fields
     * @param {?=} defaultValidatorOptions
     */
    constructor(factoryModel, fields, defaultValidatorOptions) {
        super({});
        this.factoryModel = factoryModel;
        this.fields = fields;
        this.defaultValidatorOptions = defaultValidatorOptions;
        this.customValidateErrors = new BehaviorSubject({});
        this._fb = new FormBuilder();
        /*
        const classValidators = DynamicFormGroup.getClassValidators<TModel>(
          this.factoryModel,
          this.fields,
          this.defaultValidatorOptions
        );
        const formGroup = this._fb.group(
          classValidators
        );
        Object.keys(formGroup.controls).forEach(key => {
          this.addControl(key, formGroup.controls[key]);
        });
        this.valueChanges.subscribe(data => {
          this.validate(
            undefined,
            this.defaultValidatorOptions
          );
        });*/
        this.fields = this.onlyFields(this.fields);
    }
    /**
     * @param {?} externalErrors
     * @return {?}
     */
    set externalErrors(externalErrors) {
        this._externalErrors = externalErrors;
        this.validate();
    }
    /**
     * @return {?}
     */
    get externalErrors() {
        return this._externalErrors;
    }
    /**
     * @param {?} validatorOptions
     * @return {?}
     */
    set validatorOptions(validatorOptions) {
        this._validatorOptions = validatorOptions;
        this.validate();
    }
    /**
     * @return {?}
     */
    get validatorOptions() {
        return this._validatorOptions;
    }
    /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?} fields
     * @param {?=} validatorOptions
     * @return {?}
     */
    static getClassValidators(factoryModel, fields, validatorOptions) {
        /** @type {?} */
        const allValidationMetadatas = getFromContainer(MetadataStorage).getTargetValidationMetadatas(factoryModel, '');
        /** @type {?} */
        const validationMetadatas = getFromContainer(MetadataStorage).getTargetValidationMetadatas(factoryModel, '', validatorOptions && validatorOptions.groups ? validatorOptions.groups : undefined);
        /** @type {?} */
        const formGroupFields = {};
        /** @type {?} */
        const validator = new Validator();
        Object.keys(fields)
            .filter(key => key.indexOf('__') !== 0)
            .forEach(key => {
            /** @type {?} */
            const conditionalValidations = [];
            validationMetadatas.forEach(validationMetadata => {
                if (validationMetadata.propertyName === key && validationMetadata.type === 'conditionalValidation') {
                    conditionalValidations.push(validationMetadata);
                }
            });
            /** @type {?} */
            const allNestedValidations = [];
            allValidationMetadatas.forEach(validationMetadata => {
                if (validationMetadata.propertyName === key && validationMetadata.type === 'nestedValidation') {
                    allNestedValidations.push(validationMetadata);
                }
            });
            /** @type {?} */
            const nestedValidations = [];
            validationMetadatas.forEach(validationMetadata => {
                if (validationMetadata.propertyName === key && validationMetadata.type === 'nestedValidation') {
                    nestedValidations.push(validationMetadata);
                }
            });
            /** @type {?} */
            let formGroupField = formGroupFields[key];
            if (formGroupField === undefined) {
                formGroupField = Array.isArray(fields[key]) ? fields[key] : [];
            }
            validationMetadatas.forEach(validationMetadata => {
                if (validationMetadata.propertyName === key && validationMetadata.type !== 'conditionalValidation') {
                    for (const typeKey in ValidationTypes) {
                        if (ValidationTypes.hasOwnProperty(typeKey)) {
                            if ((allNestedValidations.length === nestedValidations.length) ||
                                (fields[key] instanceof DynamicFormGroup &&
                                    (allNestedValidations.length > 0 && nestedValidations.length === 0)) ||
                                (fields[key] instanceof FormArray &&
                                    (allNestedValidations.length > 0 && nestedValidations.length === 0))) {
                                if (validationMetadata.type === ValidationTypes[typeKey] &&
                                    validator[validationMetadata.type] === undefined &&
                                    validationMetadata.type === 'nestedValidation' &&
                                    typeKey === 'NESTED_VALIDATION') {
                                    /** @type {?} */
                                    let objectToValidate;
                                    if (fields[key] instanceof DynamicFormGroup) {
                                        fields[key].object = fields[key].fields;
                                        objectToValidate = fields[key].object;
                                    }
                                    if (formGroupField.length === 0) {
                                        formGroupField.push(fields[key]);
                                    }
                                    /** @type {?} */
                                    const nestedValidate = function (c) {
                                        /** @type {?} */
                                        const isValid = c.parent && c.parent.value ? validateSync(objectToValidate || c.value, validatorOptions).length === 0 : true;
                                        return isValid
                                            ? null
                                            : {
                                                nestedValidate: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                    };
                                    formGroupField.push(nestedValidate);
                                }
                            }
                            if (validationMetadata.type === ValidationTypes[typeKey] &&
                                validator[validationMetadata.type] === undefined &&
                                validationMetadata.type === 'customValidation' &&
                                typeKey === 'CUSTOM_VALIDATION') {
                                /** @type {?} */
                                const customValidation = function (c) {
                                    // todo: refactor
                                    /** @type {?} */
                                    const object = c.parent instanceof DynamicFormGroup
                                        ? ((/** @type {?} */ (c.parent))).object
                                        : c.parent
                                            ? c.parent.value
                                            : {};
                                    if (object) {
                                        object[key] = c.value;
                                    }
                                    /** @type {?} */
                                    const validateErrors = c.parent && c.parent.value ? validateSync(object, validatorOptions) : [];
                                    /** @type {?} */
                                    const isValid = validateErrors.filter((error) => {
                                        if (error.children.length && error.children.filter(children => children.property === key)) {
                                            return true;
                                        }
                                        return true;
                                    }).length === 0;
                                    return isValid
                                        ? null
                                        : {
                                            customValidation: {
                                                valid: false,
                                                type: validationMetadata.type
                                            }
                                        };
                                };
                                if (fields[key] instanceof DynamicFormGroup) {
                                    fields[key].object = fields[key].fields;
                                }
                                if (formGroupField.length === 0) {
                                    formGroupField.push(fields[key]);
                                }
                                formGroupField.push(customValidation);
                            }
                            if (validationMetadata.type === ValidationTypes[typeKey] &&
                                validator[validationMetadata.type] !== undefined) {
                                /** @type {?} */
                                const customValidate = function (c) {
                                    if (!c) {
                                        return null;
                                    }
                                    /** @type {?} */
                                    let isValid = c.parent && c.parent.value
                                        ? validator.validateValueByMetadata(c.value, validationMetadata)
                                        : true;
                                    if (!isValid && conditionalValidations.length > 0) {
                                        // todo: refactor
                                        /** @type {?} */
                                        const object = c.parent instanceof DynamicFormGroup
                                            ? ((/** @type {?} */ (c.parent))).object
                                            : c.parent
                                                ? c.parent.value
                                                : {};
                                        if (object) {
                                            object[key] = c.value;
                                        }
                                        /** @type {?} */
                                        const validateErrors = c.parent && c.parent.value ? validateSync(object, validatorOptions) : [];
                                        isValid =
                                            validateErrors.filter((error) => {
                                                if (error.property === key) {
                                                    return true;
                                                }
                                                return false;
                                            }).length === 0;
                                    }
                                    return isValid
                                        ? null
                                        : {
                                            customValidate: {
                                                valid: false,
                                                type: validationMetadata.type
                                            }
                                        };
                                };
                                if (fields[key] instanceof DynamicFormGroup) {
                                    fields[key].object = fields[key].fields;
                                }
                                if (formGroupField.length === 0) {
                                    formGroupField.push(fields[key]);
                                }
                                formGroupField.push(customValidate);
                            }
                        }
                    }
                }
            });
            if (formGroupField[0] instanceof DynamicFormGroup) {
                formGroupFields[key] = formGroupField[0];
            }
            else {
                if (formGroupField[0] instanceof FormArray) {
                    formGroupFields[key] = formGroupField[0];
                }
                else {
                    formGroupFields[key] = [formGroupField[0], formGroupField.filter((item, index) => index !== 0)];
                }
            }
        });
        return formGroupFields;
    }
    /**
     * @private
     * @param {?} fields
     * @return {?}
     */
    onlyFields(fields) {
        /** @type {?} */
        const newFields = {};
        if (fields !== undefined) {
            Object.keys(fields).forEach(key => {
                if (fields[key] instanceof DynamicFormGroup) {
                    newFields[key] = this.onlyFields(((/** @type {?} */ (fields[key]))).fields);
                }
                else {
                    if (fields[key] instanceof FormArray) {
                        if (((/** @type {?} */ (fields[key]))).controls[0] instanceof DynamicFormGroup) {
                            newFields[key] = this.onlyFields(((/** @type {?} */ (((/** @type {?} */ (fields[key]))).controls[0]))).fields);
                        }
                        else {
                            newFields[key] = ((/** @type {?} */ (fields[key]))).controls[0].value;
                        }
                    }
                    else {
                        newFields[key] = fields[key];
                    }
                }
            });
        }
        return newFields;
    }
    /**
     * @return {?}
     */
    get object() {
        /** @type {?} */
        const object = this._object ? this.classToClass(this._object) : new this.factoryModel();
        if (object !== undefined) {
            Object.keys(this.controls).forEach(key => {
                if (this.controls[key] instanceof DynamicFormGroup) {
                    object[key] = ((/** @type {?} */ (this.controls[key]))).object;
                }
                else {
                    if (this.controls[key] instanceof FormArray) {
                        object[key] = [];
                        for (let i = 0; i < ((/** @type {?} */ (this.controls[key]))).controls.length; i++) {
                            if (((/** @type {?} */ (this.controls[key]))).controls[i] instanceof DynamicFormGroup) {
                                object[key].push(((/** @type {?} */ (((/** @type {?} */ (this.controls[key]))).controls[i]))).object);
                            }
                            else {
                                object[key].push(((/** @type {?} */ (this.controls[key]))).controls[i].value);
                            }
                        }
                    }
                    else {
                        object[key] = this.controls[key].value;
                    }
                }
            });
        }
        return this.plainToClass(this.factoryModel, object);
    }
    /**
     * @param {?} object
     * @return {?}
     */
    set object(object) {
        if (object instanceof this.factoryModel) {
            this._object = this.classToClass(object);
        }
        else {
            this._object = this.plainToClass(this.factoryModel, (/** @type {?} */ (object)));
        }
        Object.keys(this.controls).forEach(key => {
            if (this.controls[key] instanceof DynamicFormGroup) {
                ((/** @type {?} */ (this.controls[key]))).object = this._object ? this._object[key] : {};
            }
            else {
                if (this.controls[key] instanceof FormArray) {
                    /** @type {?} */
                    const objectArray = this._object ? this._object[key] : [];
                    /** @type {?} */
                    const formArray = ((/** @type {?} */ (this.controls[key])));
                    /** @type {?} */
                    const isFormGroup = formArray.controls[0] instanceof DynamicFormGroup;
                    /** @type {?} */
                    const prevFormGroup = (/** @type {?} */ (formArray.controls[0]));
                    /** @type {?} */
                    const formControl = (/** @type {?} */ (formArray.controls[0]));
                    while (formArray.length !== 0) {
                        formArray.removeAt(0);
                    }
                    for (let i = 0; i < objectArray.length; i++) {
                        if (isFormGroup) {
                            /** @type {?} */
                            const dynamicFormGroup = new DynamicFormGroup(prevFormGroup.factoryModel, prevFormGroup.fields, this._validatorOptions);
                            dynamicFormGroup.setParent(this);
                            /** @type {?} */
                            const classValidators = DynamicFormGroup.getClassValidators(prevFormGroup.factoryModel, prevFormGroup.fields, this._validatorOptions);
                            /** @type {?} */
                            const formGroup = this._fb.group(classValidators);
                            Object.keys(formGroup.controls).forEach(ctrlKey => {
                                dynamicFormGroup.addControl(ctrlKey, formGroup.controls[ctrlKey]);
                            });
                            dynamicFormGroup.valueChanges.subscribe(data => {
                                dynamicFormGroup.validate(undefined, this._validatorOptions);
                            });
                            formArray.controls.push(dynamicFormGroup);
                            ((/** @type {?} */ (formArray.controls[i]))).object =
                                (this._object && objectArray && objectArray[i]) ? objectArray[i] : {};
                        }
                        else {
                            /** @type {?} */
                            const newFormControl = new FormControl((this._object && objectArray && objectArray[i]) ? objectArray[i] : undefined, formControl.validator);
                            newFormControl.setParent(this);
                            formArray.controls.push(newFormControl);
                        }
                    }
                }
                else {
                    /** @type {?} */
                    const newObject = this._object ? this._object[key] : [];
                    this.controls[key].setValue(this._object && newObject ? newObject : undefined);
                }
            }
        });
    }
    /**
     * @private
     * @param {?} errors
     * @return {?}
     */
    transformValidationErrors(errors) {
        /** @type {?} */
        const customErrors = {};
        errors.forEach((error) => {
            if (error && error.constraints !== undefined) {
                Object.keys(error.constraints).forEach((key) => {
                    if (!customErrors[error.property]) {
                        customErrors[error.property] = [];
                    }
                    if (((/** @type {?} */ (customErrors[error.property]))).indexOf(error.constraints[key]) === -1) {
                        ((/** @type {?} */ (customErrors[error.property]))).push(error.constraints[key]);
                    }
                });
            }
            if (error.children !== undefined && error.children.length) {
                customErrors[error.property] = this.transformValidationErrors(error.children);
                if (Array.isArray(error.value)) {
                    customErrors[error.property] = Object.keys(customErrors[error.property]).map(key => customErrors[error.property][key]);
                }
            }
        });
        return customErrors;
    }
    /**
     * @template TClassModel
     * @param {?} object
     * @return {?}
     */
    classToClass(object) {
        return classToClass(object, { ignoreDecorators: true });
    }
    /**
     * @template TClassModel, Object
     * @param {?} cls
     * @param {?} plain
     * @return {?}
     */
    plainToClass(cls, plain) {
        return plainToClass(cls, plain, { ignoreDecorators: true });
    }
    /**
     * @private
     * @param {?=} errors
     * @param {?=} externalErrors
     * @return {?}
     */
    mergeErrors(errors, externalErrors) {
        return mergeWith(errors, externalErrors, (objValue, srcValue) => {
            if (Array.isArray(objValue) &&
                Array.isArray(srcValue) &&
                objValue.filter(objItem => srcValue.indexOf(objItem) !== -1).length === 0) {
                return objValue.concat(srcValue);
            }
        });
    }
    /**
     * @param {?=} externalErrors
     * @param {?=} validatorOptions
     * @return {?}
     */
    validateAsync(externalErrors, validatorOptions) {
        return __awaiter(this, void 0, void 0, function* () {
            if (externalErrors === undefined) {
                externalErrors = cloneDeep(this.externalErrors);
            }
            if (validatorOptions === undefined) {
                validatorOptions = cloneDeep(this.validatorOptions);
            }
            if (!externalErrors) {
                externalErrors = {};
            }
            try {
                /** @type {?} */
                const result = yield validate(this.object, validatorOptions);
                /** @type {?} */
                const transformedErrors = this.transformValidationErrors(result);
                /** @type {?} */
                const allErrors = this.mergeErrors(externalErrors, transformedErrors);
                this.markAsInvalidForExternalErrors(externalErrors, this.controls);
                this.customValidateErrors.next(allErrors);
            }
            catch (error) {
                throw error;
            }
        });
    }
    /**
     * @param {?=} externalErrors
     * @param {?=} validatorOptions
     * @return {?}
     */
    validate(externalErrors, validatorOptions) {
        this.validateAsync(externalErrors, validatorOptions).then(() => { }, error => { throw error; });
    }
    /**
     * @private
     * @param {?} errors
     * @param {?} controls
     * @return {?}
     */
    markAsInvalidForExternalErrors(errors, controls) {
        Object.keys(controls).forEach(field => {
            /** @type {?} */
            const control = controls[field];
            if (control instanceof FormControl) {
                if (errors && errors[field]) {
                    control.setErrors({ externalError: true });
                }
                else {
                    if (control.errors && control.errors.externalError === true) {
                        control.setErrors(null);
                    }
                }
            }
            else {
                if (control instanceof DynamicFormGroup) {
                    control.markAsInvalidForExternalErrors(errors && errors[field] ? ((/** @type {?} */ (errors[field]))) : {}, control.controls);
                }
                else {
                    if (control instanceof FormArray) {
                        for (let i = 0; i < ((/** @type {?} */ (control))).length; i++) {
                            if (control[i] instanceof FormControl) {
                                if (errors && errors[i] && errors[i][field]) {
                                    control[i].setErrors({ externalError: true });
                                }
                                else {
                                    if (control[i].errors && control[i].errors.externalError === true) {
                                        control[i].setErrors(null);
                                    }
                                }
                            }
                            else {
                                if (control[i] instanceof DynamicFormGroup) {
                                    control[i].markAsInvalidForExternalErrors(errors && errors[i] && errors[i][field] ? ((/** @type {?} */ (errors[i][field]))) : {}, control[i].controls);
                                }
                            }
                        }
                    }
                }
            }
        });
    }
    /**
     * @return {?}
     */
    validateAllFormFields() {
        Object.keys(this.controls).forEach(field => {
            /** @type {?} */
            const control = this.get(field);
            if (control instanceof FormControl) {
                control.markAsTouched({ onlySelf: true });
            }
            else {
                if (control instanceof DynamicFormGroup) {
                    control.validateAllFormFields();
                }
                else {
                    if (control instanceof FormArray) {
                        for (let i = 0; i < ((/** @type {?} */ (control))).controls.length; i++) {
                            if (((/** @type {?} */ (control))).controls[i] instanceof FormControl) {
                                ((/** @type {?} */ (((/** @type {?} */ (control))).controls[i]))).markAsTouched({ onlySelf: true });
                            }
                            else {
                                if (((/** @type {?} */ (control))).controls[i] instanceof DynamicFormGroup) {
                                    ((/** @type {?} */ (((/** @type {?} */ (control))).controls[i]))).validateAllFormFields();
                                }
                            }
                        }
                    }
                }
            }
        });
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class DynamicFormBuilder extends FormBuilder {
    /**
     * @private
     * @template TModel
     * @param {?} factoryModel
     * @param {?=} data
     * @return {?}
     */
    createEmptyObject(factoryModel, data = {}) {
        /** @type {?} */
        const object = factoryModel ? plainToClass(factoryModel, data) : data;
        /** @type {?} */
        let modifed = false;
        /** @type {?} */
        const keys = Object.keys(object);
        keys.forEach((key) => {
            if (object[key] && object[key].length !== undefined) {
                if (object[key].length === 1 && Object.keys(object[key][0]).length > 0 && object[key][0].constructor) {
                    object[key] = [this.createEmptyObject(object[key][0].constructor)];
                }
                if (object[key].length === 0) {
                    data[key] = [
                        {}
                    ];
                    modifed = true;
                }
            }
            else {
                data[key] = undefined;
            }
        });
        if (modifed) {
            return this.createEmptyObject(factoryModel, data);
        }
        return object;
    }
    /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?=} controlsConfig
     * @param {?=} extra
     * @return {?}
     */
    group(factoryModel, controlsConfig, extra) {
        if (controlsConfig && (controlsConfig.legacyOrOpts || controlsConfig.customValidatorOptions)) {
            return this.group(factoryModel, undefined, controlsConfig);
        }
        if (extra !== undefined && extra.customValidatorOptions === undefined) {
            extra.customValidatorOptions = { validationError: { target: false } };
        }
        /** @type {?} */
        let newControlsConfig;
        if (controlsConfig !== undefined) {
            newControlsConfig = controlsConfig;
        }
        // experimental
        if (controlsConfig === undefined) {
            newControlsConfig = Object.assign({}, this.createEmptyObject(factoryModel));
            Object.keys(newControlsConfig).forEach(key => {
                if (newControlsConfig[key] &&
                    !Array.isArray(newControlsConfig[key]) &&
                    newControlsConfig[key].constructor &&
                    typeof newControlsConfig[key] === 'object' &&
                    (newControlsConfig[key].length === undefined ||
                        (newControlsConfig[key].length !== undefined &&
                            Object.keys(newControlsConfig[key]).length === newControlsConfig[key].length))) {
                    newControlsConfig[key] = this.group(newControlsConfig[key].constructor, extra);
                }
                else {
                    if (Array.isArray(newControlsConfig[key]) &&
                        newControlsConfig[key][0].constructor &&
                        typeof newControlsConfig[key][0] === 'object' &&
                        (newControlsConfig[key][0].length === undefined ||
                            (newControlsConfig[key][0].length !== undefined &&
                                Object.keys(newControlsConfig[key][0]).length === newControlsConfig[key][0].length))) {
                        if (newControlsConfig[key][0].constructor) {
                            newControlsConfig[key] = super.array(newControlsConfig[key].map(newControlsConfigItem => this.group(newControlsConfigItem.constructor, extra)));
                        }
                        else {
                            newControlsConfig[key] = super.array(newControlsConfig[key].map(newControlsConfigItem => this.control(newControlsConfigItem)));
                        }
                    }
                }
            });
        }
        /** @type {?} */
        const dynamicFormGroup = new DynamicFormGroup(factoryModel, newControlsConfig, extra && extra.customValidatorOptions ? extra.customValidatorOptions : undefined);
        /** @type {?} */
        const classValidators = DynamicFormGroup.getClassValidators(factoryModel, newControlsConfig, extra && extra.customValidatorOptions ? extra.customValidatorOptions : undefined);
        /** @type {?} */
        const formGroup = super.group(classValidators, extra);
        Object.keys(formGroup.controls).forEach(key => {
            dynamicFormGroup.addControl(key, formGroup.controls[key]);
        });
        dynamicFormGroup.valueChanges.subscribe(data => {
            dynamicFormGroup.validate(undefined, extra && extra.customValidatorOptions ? extra.customValidatorOptions : undefined);
        });
        return dynamicFormGroup;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { DynamicFormBuilder, DynamicFormGroup };

//# sourceMappingURL=ngx-dynamic-form-builder.js.map